package JavaScript::Engine;

our $VERSION = '0.014';


=head1 NAME

JavaScript::Engine - Pure-Perl ECMAScript (JavaScript) engine

=head1 DESCRIPTION

This is just a pointer to L<JE>, a pure-Perl JS engine.

"JE" stands for 
JavaScript::Engine (following the example of PPI). I included this module
in the JE distribution so that
one might be able to find JE when typing C<i /javascript/> in the CPAN
shell or when looking under modules/by-module/JavaScript/ on the CPAN.

=head1 AUTHOR

Father Chrysostomos <sprout at cpan.org>

=cut

1; # End of JavaScript::Engine
